@@ -40,7 +40,9 @@ MVMObject * MVM_sc_create(MVMThreadContext *tc, MVMString *handle) {
                 scb->sc = sc;
                 sc->body = scb;
                 MVM_ASSIGN_REF(tc, &(sc->common.header), scb->handle, handle);
+                MVM_gc_allocate_gen2_default_set(tc);
                 MVM_repr_init(tc, (MVMObject *)sc);
+                MVM_gc_allocate_gen2_default_clear(tc);
             }
             uv_mutex_unlock(&tc->instance->mutex_sc_weakhash);
         });
@@ -102,6 +104,14 @@ MVMint64 MVM_sc_find_object_idx(MVMThreadContext *tc, MVMSerializationContext *s
         "Object does not exist in serialization context");
 }
 
+/* Calls MVM_sc_find_object_idx, but first checks if the sc is actually an SCRef. */
+MVMint64 MVM_sc_find_object_idx_jit(MVMThreadContext *tc, MVMObject *sc, MVMObject *obj) {
+    if (REPR(sc)->ID != MVM_REPR_ID_SCRef)
+        MVM_exception_throw_adhoc(tc,
+            "Must provide an SCRef operand to scgetobjidx");
+    return MVM_sc_find_object_idx(tc, (MVMSerializationContext *)sc, obj);
+}
+
 /* Given an SC, looks up the index of an STable that is in its root set. */
 MVMint64 MVM_sc_find_stable_idx(MVMThreadContext *tc, MVMSerializationContext *sc, MVMSTable *st) {
     MVMuint64 i;
@@ -152,6 +162,7 @@ MVMSerializationContext * MVM_sc_get_sc(MVMThreadContext *tc, MVMCompUnit *cu, M
         if (sc == NULL)
             return NULL;
         MVM_ASSIGN_REF(tc, &(cu->common.header), cu->body.scs[dep], sc);
+        scb->claimed = 1;
     }
     return sc;
 }
