@@ -62,6 +62,9 @@ typedef struct {
     /* The limit we can not read beyond. */
     MVMuint8 *read_limit;
 
+    /* Array of frames. */
+    MVMStaticFrame **frames;
+
     /* Special frame indexes */
     MVMuint32  main_frame;
     MVMuint32  load_frame;
@@ -102,10 +105,8 @@ static MVMuint8 read_int8(MVMuint8 *buffer, size_t offset) {
 
 /* Cleans up reader state. */
 static void cleanup_all(MVMThreadContext *tc, ReaderState *rs) {
-    if (rs->frame_outer_fixups) {
-        MVM_free(rs->frame_outer_fixups);
-        rs->frame_outer_fixups = NULL;
-    }
+    MVM_free(rs->frames);
+    MVM_free(rs->frame_outer_fixups);
     MVM_free(rs);
 }
 
@@ -121,14 +122,13 @@ MVM_STATIC_INLINE void ensure_can_read(MVMThreadContext *tc, MVMCompUnit *cu, Re
 /* Reads a string index, looks up the string and returns it. Bounds
  * checks the string heap index too. */
 static MVMString * get_heap_string(MVMThreadContext *tc, MVMCompUnit *cu, ReaderState *rs, MVMuint8 *buffer, size_t offset) {
-    MVMCompUnitBody *cu_body = &cu->body;
     MVMuint32 heap_index = read_int32(buffer, offset);
-    if (heap_index >= cu_body->num_strings) {
+    if (heap_index >= cu->body.num_strings) {
         if (rs)
             cleanup_all(tc, rs);
         MVM_exception_throw_adhoc(tc, "String heap index beyond end of string heap");
     }
-    return cu_body->strings[heap_index];
+    return MVM_cu_string(tc, cu, heap_index);
 }
 
 /* Dissects the bytecode stream and hands back a reader pointing to the
@@ -250,49 +250,6 @@ static ReaderState * dissect_bytecode(MVMThreadContext *tc, MVMCompUnit *cu) {
     return rs;
 }
 
-/* Loads the string heap. */
-static MVMString ** deserialize_strings(MVMThreadContext *tc, MVMCompUnit *cu, ReaderState *rs) {
-    MVMString **strings;
-    MVMuint8   *pos;
-    MVMuint32   i, ss;
-
-    /* Allocate space for strings list. */
-    if (rs->expected_strings == 0)
-        return NULL;
-    strings = MVM_malloc(sizeof(MVMString *) * rs->expected_strings);
-
-    /* Load strings. */
-    pos = rs->string_seg;
-    for (i = 0; i < rs->expected_strings; i++) {
-        MVMint32 decode_utf8 = 1;
-
-        /* Ensure we can read at least a string size here and do so. */
-        ensure_can_read(tc, cu, rs, pos, 4);
-        ss = read_int32(pos, 0);
-        pos += 4;
-
-        /* At high enough bytecode file versions, the LSB on this number
-         * is 1 if we should decode as UTF-8, and 0 if Latin-1 will do. */
-        if (rs->version >= 5) {
-            decode_utf8 = ss & 1;
-            ss = ss >> 1;
-        }
-
-        /* Ensure we can read in the string of this size, and decode
-         * it if so. */
-        ensure_can_read(tc, cu, rs, pos, ss);
-        MVM_ASSIGN_REF(tc, &(cu->common.header), strings[i], decode_utf8
-            ? MVM_string_utf8_decode(tc, tc->instance->VMString, (char *)pos, ss)
-            : MVM_string_latin1_decode(tc, tc->instance->VMString, (char *)pos, ss));
-        pos += ss;
-
-        /* Add alignment. */
-        pos += ss & 3 ? 4 - (ss & 3) : 0;
-    }
-
-    return strings;
-}
-
 /* Loads the SC dependencies list. */
 static void deserialize_sc_deps(MVMThreadContext *tc, MVMCompUnit *cu, ReaderState *rs) {
     MVMCompUnitBody *cu_body = &cu->body;
@@ -322,7 +279,7 @@ static void deserialize_sc_deps(MVMThreadContext *tc, MVMCompUnit *cu, ReaderSta
             MVM_exception_throw_adhoc(tc, "String heap index beyond end of string heap");
         }
         cu_body->sc_handle_idxs[i] = sh_idx;
-        handle = cu_body->strings[sh_idx];
+        handle = MVM_cu_string(tc, cu, sh_idx);
 
         /* See if we can resolve it. */
         uv_mutex_lock(&tc->instance->mutex_sc_weakhash);
@@ -375,7 +332,7 @@ static MVMExtOpRecord * deserialize_extop_records(MVMThreadContext *tc, MVMCompU
             MVM_exception_throw_adhoc(tc,
                     "String heap index beyond end of string heap");
         }
-        extops[i].name = cu->body.strings[name_idx];
+        extops[i].name = MVM_cu_string(tc, cu, name_idx);
 
         /* Read operand descriptor. */
         ensure_can_read(tc, cu, rs, pos, 8);
@@ -462,6 +419,10 @@ static MVMExtOpRecord * deserialize_extop_records(MVMThreadContext *tc, MVMCompU
                     case MVM_operand_str:
                     case MVM_operand_obj:
                     case MVM_operand_type_var:
+                    case MVM_operand_uint8:
+                    case MVM_operand_uint16:
+                    case MVM_operand_uint32:
+                    case MVM_operand_uint64:
                         continue;
 
                     default:
@@ -481,7 +442,7 @@ static MVMExtOpRecord * deserialize_extop_records(MVMThreadContext *tc, MVMCompU
 }
 
 /* Loads the static frame information (what locals we have, bytecode offset,
- * lexicals, etc.) */
+ * lexicals, etc.) and returns a list of them. */
 static MVMStaticFrame ** deserialize_frames(MVMThreadContext *tc, MVMCompUnit *cu, ReaderState *rs) {
     MVMStaticFrame **frames;
     MVMuint8        *pos;
@@ -778,6 +739,8 @@ static MVMCallsite ** deserialize_callsites(MVMThreadContext *tc, MVMCompUnit *c
         callsites[i]->flag_count = elems;
         if (elems)
             callsites[i]->arg_flags = MVM_malloc(elems);
+        else
+            callsites[i]->arg_flags = NULL;
 
         /* Ensure we can read in a callsite of this size, and do so. */
         ensure_can_read(tc, cu, rs, pos, elems);
@@ -853,7 +816,7 @@ static MVMCallsite ** deserialize_callsites(MVMThreadContext *tc, MVMCompUnit *c
 }
 
 /* Creates code objects to go with each of the static frames. */
-static void create_code_objects(MVMThreadContext *tc, MVMCompUnit *cu) {
+static void create_code_objects(MVMThreadContext *tc, MVMCompUnit *cu, ReaderState *rs) {
     MVMuint32  i;
     MVMObject *code_type;
     MVMCompUnitBody *cu_body = &cu->body;
@@ -864,9 +827,9 @@ static void create_code_objects(MVMThreadContext *tc, MVMCompUnit *cu) {
     for (i = 0; i < cu_body->num_frames; i++) {
         MVMCode *coderef = (MVMCode *)REPR(code_type)->allocate(tc, STABLE(code_type));
         MVM_ASSIGN_REF(tc, &(cu->common.header), cu_body->coderefs[i], coderef);
-        MVM_ASSIGN_REF(tc, &(coderef->common.header), coderef->body.sf, cu_body->frames[i]);
-        MVM_ASSIGN_REF(tc, &(coderef->common.header), coderef->body.name, cu_body->frames[i]->body.name);
-        MVM_ASSIGN_REF(tc, &(cu_body->frames[i]->common.header), cu_body->frames[i]->body.static_code, coderef);
+        MVM_ASSIGN_REF(tc, &(coderef->common.header), coderef->body.sf, rs->frames[i]);
+        MVM_ASSIGN_REF(tc, &(coderef->common.header), coderef->body.name, rs->frames[i]->body.name);
+        MVM_ASSIGN_REF(tc, &(rs->frames[i]->common.header), rs->frames[i]->body.static_code, coderef);
     }
 }
 
@@ -882,10 +845,15 @@ void MVM_bytecode_unpack(MVMThreadContext *tc, MVMCompUnit *cu) {
     /* Dissect the bytecode into its parts. */
     rs = dissect_bytecode(tc, cu);
 
-    /* Load the strings heap. */
-    cu_body->strings = deserialize_strings(tc, cu, rs);
+    /* Allocate space for the strings heap; we deserialize it lazily. */
+    cu_body->strings = MVM_calloc(rs->expected_strings, sizeof(MVMString *));
     cu_body->num_strings = rs->expected_strings;
     cu_body->orig_strings = rs->expected_strings;
+    cu_body->string_heap_fast_table = MVM_calloc(
+        (rs->expected_strings / MVM_STRING_FAST_TABLE_SPAN) + 1,
+        sizeof(MVMuint32));
+    cu_body->string_heap_start = rs->string_seg;
+    cu_body->string_heap_read_limit = rs->read_limit;
 
     /* Load SC dependencies. */
     deserialize_sc_deps(tc, cu, rs);
@@ -895,10 +863,10 @@ void MVM_bytecode_unpack(MVMThreadContext *tc, MVMCompUnit *cu) {
     cu_body->num_extops = rs->expected_extops;
 
     /* Load the static frame info and give each one a code reference. */
-    cu_body->frames = deserialize_frames(tc, cu, rs);
+    rs->frames = deserialize_frames(tc, cu, rs);
     cu_body->num_frames = rs->expected_frames;
     cu_body->orig_frames = rs->expected_frames;
-    create_code_objects(tc, cu);
+    create_code_objects(tc, cu, rs);
 
     /* Load callsites. */
     cu_body->max_callsite_size = MVM_MIN_CALLSITE_SIZE;
@@ -906,16 +874,20 @@ void MVM_bytecode_unpack(MVMThreadContext *tc, MVMCompUnit *cu) {
     cu_body->num_callsites = rs->expected_callsites;
     cu_body->orig_callsites = rs->expected_callsites;
 
+    if (rs->hll_str_idx > rs->expected_strings)
+        MVM_exception_throw_adhoc(tc, "Unpacking bytecode: HLL name string index out of range: %d > %d", rs->hll_str_idx, rs->expected_strings);
+
     /* Resolve HLL name. */
-    MVM_ASSIGN_REF(tc, &(cu->common.header), cu_body->hll_name, cu_body->strings[rs->hll_str_idx]);
+    MVM_ASSIGN_REF(tc, &(cu->common.header), cu_body->hll_name,
+        MVM_cu_string(tc, cu, rs->hll_str_idx));
 
     /* Resolve special frames. */
-    if (rs->main_frame)
-        MVM_ASSIGN_REF(tc, &(cu->common.header), cu_body->main_frame, cu_body->frames[rs->main_frame - 1]);
+    MVM_ASSIGN_REF(tc, &(cu->common.header), cu_body->main_frame,
+        rs->main_frame ? rs->frames[rs->main_frame - 1] : rs->frames[0]);
     if (rs->load_frame)
-        MVM_ASSIGN_REF(tc, &(cu->common.header), cu_body->load_frame, cu_body->frames[rs->load_frame - 1]);
+        MVM_ASSIGN_REF(tc, &(cu->common.header), cu_body->load_frame, rs->frames[rs->load_frame - 1]);
     if (rs->deserialize_frame)
-        MVM_ASSIGN_REF(tc, &(cu->common.header), cu_body->deserialize_frame, cu_body->frames[rs->deserialize_frame - 1]);
+        MVM_ASSIGN_REF(tc, &(cu->common.header), cu_body->deserialize_frame, rs->frames[rs->deserialize_frame - 1]);
 
     /* Clean up reader state. */
     cleanup_all(tc, rs);
