@@ -31,34 +31,37 @@ MVMHLLConfig *MVM_hll_get_config_for(MVMThreadContext *tc, MVMString *name) {
             HASH_ADD_KEYPTR(hash_handle, tc->instance->compilee_hll_configs, kdata, klen, entry);
         else
             HASH_ADD_KEYPTR(hash_handle, tc->instance->compiler_hll_configs, kdata, klen, entry);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->int_box_type);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->num_box_type);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->str_box_type);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->slurpy_array_type);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->slurpy_hash_type);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->array_iterator_type);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->hash_iterator_type);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->foreign_type_int);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->foreign_type_num);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->foreign_type_str);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->foreign_transform_array);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->foreign_transform_hash);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->foreign_transform_code);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->null_value);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->exit_handler);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->finalize_handler);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->bind_error);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->method_not_found_error);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->int_lex_ref);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->num_lex_ref);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->str_lex_ref);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->int_attr_ref);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->num_attr_ref);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->str_attr_ref);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->int_pos_ref);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->num_pos_ref);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->str_pos_ref);
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->name);
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->int_box_type, "HLL int_box_type");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->num_box_type, "HLL num_box_type");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->str_box_type, "HLL str_box_type");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->slurpy_array_type, "HLL slurpy_array_type");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->slurpy_hash_type, "HLL slurpy_hash_type");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->array_iterator_type, "HLL array_iterator_type");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->hash_iterator_type, "HLL hash_iterator_type");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->foreign_type_int, "HLL foreign_type_int");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->foreign_type_num, "HLL foreign_type_num");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->foreign_type_str, "HLL foreign_type_str");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->foreign_transform_array, "HLL foreign_transform_array");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->foreign_transform_hash, "HLL foreign_transform_hash");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->foreign_transform_code, "HLL foreign_transform_code");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->null_value, "HLL null_value");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->exit_handler, "HLL exit_handler");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->finalize_handler, "HLL finalize_handler");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->bind_error, "HLL bind_error");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->method_not_found_error, "HLL method_not_found_error");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->int_lex_ref, "HLL int_lex_ref");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->num_lex_ref, "HLL num_lex_ref");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->str_lex_ref, "HLL str_lex_ref");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->int_attr_ref, "HLL int_attr_ref");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->num_attr_ref, "HLL num_attr_ref");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->str_attr_ref, "HLL str_attr_ref");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->int_pos_ref, "HLL int_pos_ref");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->num_pos_ref, "HLL num_pos_ref");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->str_pos_ref, "HLL str_pos_ref");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->int_multidim_ref, "HLL int_multidim_ref");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->num_multidim_ref, "HLL num_multidim_ref");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->str_multidim_ref, "HLL str_multidim_ref");
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->name, "HLL name");
     }
 
     uv_mutex_unlock(&tc->instance->mutex_hllconfigs);
@@ -128,6 +131,12 @@ MVMObject * MVM_hll_set_config(MVMThreadContext *tc, MVMString *name, MVMObject
                 config, MVM_STORAGE_SPEC_BP_NUM, MVM_NATIVEREF_POSITIONAL);
             check_config_key_reftype(tc, config_hash, "str_pos_ref", str_pos_ref,
                 config, MVM_STORAGE_SPEC_BP_STR, MVM_NATIVEREF_POSITIONAL);
+            check_config_key_reftype(tc, config_hash, "int_multidim_ref", int_multidim_ref,
+                config, MVM_STORAGE_SPEC_BP_INT, MVM_NATIVEREF_MULTIDIM);
+            check_config_key_reftype(tc, config_hash, "num_multidim_ref", num_multidim_ref,
+                config, MVM_STORAGE_SPEC_BP_NUM, MVM_NATIVEREF_MULTIDIM);
+            check_config_key_reftype(tc, config_hash, "str_multidim_ref", str_multidim_ref,
+                config, MVM_STORAGE_SPEC_BP_STR, MVM_NATIVEREF_MULTIDIM);
         });
 
     MVM_intcache_for(tc, config->int_box_type);
