@@ -184,6 +184,8 @@ static void finish_gc(MVMThreadContext *tc, MVMuint8 gen, MVMuint8 is_coordinato
             "Thread %d run %d : Co-ordinator handling fixed-size allocator safepoint frees\n");
         MVM_fixed_size_safepoint(tc, tc->instance->fsa);
 
+        MVM_profile_heap_take_snapshot(tc);
+
         GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
             "Thread %d run %d : Co-ordinator signalling in-trays clear\n");
         MVM_store(&tc->instance->gc_intrays_clearing, 0);
@@ -270,9 +272,19 @@ void MVM_gc_mark_thread_unblocked(MVMThreadContext *tc) {
 }
 
 static MVMint32 is_full_collection(MVMThreadContext *tc) {
-    MVMuint64 threshold = MVM_GC_GEN2_THRESHOLD_BASE +
-        ((MVMuint64)tc->instance->num_user_threads * MVM_GC_GEN2_THRESHOLD_THREAD);
-    return MVM_load(&tc->instance->gc_promoted_bytes_since_last_full) > threshold;
+    MVMuint64 percent_growth, threshold;
+    size_t rss, promoted;
+
+    /* If it's below the absolute minimum, quickly return. */
+    promoted = (size_t)MVM_load(&tc->instance->gc_promoted_bytes_since_last_full);
+    if (promoted < MVM_GC_GEN2_THRESHOLD_MINIMUM)
+        return 0;
+
+    /* Otherwise, consider percentage of resident set size. */
+    if (uv_resident_set_memory(&rss) < 0)
+        rss = 50 * 1024 * 1024;
+    percent_growth = (100 * promoted) / rss;
+    return percent_growth >= MVM_GC_GEN2_THRESHOLD_PERCENT;
 }
 
 static void run_gc(MVMThreadContext *tc, MVMuint8 what_to_do) {
@@ -280,7 +292,7 @@ static void run_gc(MVMThreadContext *tc, MVMuint8 what_to_do) {
     MVMuint32  i, n;
 
     /* Decide nursery or full collection. */
-    gen = is_full_collection(tc) ? MVMGCGenerations_Both : MVMGCGenerations_Nursery;
+    gen = tc->instance->gc_full_collect ? MVMGCGenerations_Both : MVMGCGenerations_Nursery;
 
     /* Do GC work for ourselves and any work threads. */
     for (i = 0, n = tc->gc_work_count ; i < n; i++) {
@@ -334,7 +346,6 @@ void MVM_gc_enter_from_allocator(MVMThreadContext *tc) {
     if (MVM_trycas(&tc->instance->gc_start, 0, 1)) {
         MVMThread *last_starter = NULL;
         MVMuint32 num_threads = 0;
-        MVMuint32 is_full;
 
         /* Need to wait for other threads to reset their gc_status. */
         while (MVM_load(&tc->instance->gc_ack)) {
@@ -352,11 +363,11 @@ void MVM_gc_enter_from_allocator(MVMThreadContext *tc) {
             (int)MVM_load(&tc->instance->gc_seq_number));
 
         /* Decide if it will be a full collection. */
-        is_full = is_full_collection(tc);
+        tc->instance->gc_full_collect = is_full_collection(tc);
 
         /* If profiling, record that GC is starting. */
         if (tc->instance->profiling)
-            MVM_profiler_log_gc_start(tc, is_full);
+            MVM_profiler_log_gc_start(tc, tc->instance->gc_full_collect);
 
         /* Ensure our stolen list is empty. */
         tc->gc_work_count = 0;
@@ -407,7 +418,7 @@ void MVM_gc_enter_from_allocator(MVMThreadContext *tc) {
 
         /* Now we're ready to start, zero promoted since last full collection
          * counter if this is a full collect. */
-        if (is_full)
+        if (tc->instance->gc_full_collect)
             MVM_store(&tc->instance->gc_promoted_bytes_since_last_full, 0);
 
         /* This is a safe point for us to free any STables that have been marked
