@@ -1,6 +1,16 @@
 #include "moar.h"
 #include <platform/threads.h>
 
+#if defined(_MSC_VER)
+#define snprintf _snprintf
+#endif
+
+#ifndef _WIN32
+#  include <unistd.h>
+#else
+#  include <process.h>
+#endif
+
 #define init_mutex(loc, name) do { \
     if ((init_stat = uv_mutex_init(&loc)) < 0) { \
         fprintf(stderr, "MoarVM: Initialization of " name " mutex failed\n    %s\n", \
@@ -11,6 +21,48 @@
 
 static void setup_std_handles(MVMThreadContext *tc);
 
+static FILE *fopen_perhaps_with_pid(char *path, const char *mode) {
+    if (strstr(path, "%d")) {
+        MVMuint64 path_length = strlen(path);
+        MVMuint64 found_percents = 0;
+        MVMuint64 i;
+
+        /* Let's sanitize the format string a bit. Must only have
+         * a single printf-recognized directive. */
+        for (i = 0; i < path_length; i++) {
+            if (path[i] == '%') {
+                /* %% is all right. */
+                if (i + 1 < path_length && path[i + 1] == '%') {
+                    i++; continue;
+                }
+                found_percents++;
+            }
+        }
+        /* We expect to pass only a single argument to snprintf here;
+         * just bail out if there's more than one directive. */
+        if (found_percents > 1) {
+            return fopen(path, mode);
+        } else {
+            char *fixed_path = malloc(path_length + 16);
+            FILE *result;
+            MVMint64 pid;
+#ifdef _WIN32
+            pid = _getpid();
+#else
+            pid = getpid();
+#endif
+            /* We make the brave assumption that
+             * pids only go up to 16 characters. */
+            snprintf(fixed_path, path_length + 16, path, pid);
+            result = fopen(fixed_path, mode);
+            free(fixed_path);
+            return result;
+        }
+    } else {
+        return fopen(path, mode);
+    }
+}
+
 /* Create a new instance of the VM. */
 MVMInstance * MVM_vm_create_instance(void) {
     MVMInstance *instance;
@@ -32,9 +84,10 @@ MVMInstance * MVM_vm_create_instance(void) {
     MVM_store(&instance->next_user_thread_id, 2);
 
     /* Set up the permanent roots storage. */
-    instance->num_permroots   = 0;
-    instance->alloc_permroots = 16;
-    instance->permroots       = MVM_malloc(sizeof(MVMCollectable **) * instance->alloc_permroots);
+    instance->num_permroots         = 0;
+    instance->alloc_permroots       = 16;
+    instance->permroots             = MVM_malloc(sizeof(MVMCollectable **) * instance->alloc_permroots);
+    instance->permroot_descriptions = MVM_malloc(sizeof(char *) * instance->alloc_permroots);
     init_mutex(instance->mutex_permroots, "permanent roots");
 
     /* Create fixed size allocator. */
@@ -71,8 +124,10 @@ MVMInstance * MVM_vm_create_instance(void) {
      * they will have program lifetime. */
     MVM_gc_allocate_gen2_default_set(instance->main_thread);
 
+    /* Set up integer constant and string cache. */
     init_mutex(instance->mutex_int_const_cache, "int constant cache");
     instance->int_const_cache = MVM_calloc(1, sizeof(MVMIntConstCache));
+    instance->int_to_str_cache = MVM_calloc(MVM_INT_TO_STR_CACHE_SIZE, sizeof(MVMString *));
 
     /* Bootstrap 6model. It is assumed the GC will not be called during this. */
     MVM_6model_bootstrap(instance->main_thread);
@@ -106,6 +161,9 @@ MVMInstance * MVM_vm_create_instance(void) {
     /* Set up hll symbol tables mutex. */
     init_mutex(instance->mutex_hll_syms, "hll syms");
 
+    /* Initialize Unicode database */
+    MVM_unicode_init(instance->main_thread);
+
     /* Initialize string cclass handling. */
     MVM_string_cclass_init(instance->main_thread);
 
@@ -113,9 +171,6 @@ MVMInstance * MVM_vm_create_instance(void) {
     instance->callsite_interns = MVM_calloc(1, sizeof(MVMCallsiteInterns));
     init_mutex(instance->mutex_callsite_interns, "callsite interns");
 
-    /* Allocate int to str cache. */
-    instance->int_to_str_cache = MVM_calloc(MVM_INT_TO_STR_CACHE_SIZE, sizeof(MVMString *));
-
     /* There's some callsites we statically use all over the place. Intern
      * them, so that spesh may end up optimizing more "internal" stuff. */
     MVM_callsite_initialize_common(instance->main_thread);
@@ -132,7 +187,7 @@ MVMInstance * MVM_vm_create_instance(void) {
     init_mutex(instance->mutex_spesh_install, "spesh installations");
     spesh_log = getenv("MVM_SPESH_LOG");
     if (spesh_log && strlen(spesh_log))
-        instance->spesh_log_fh = fopen(spesh_log, "w");
+        instance->spesh_log_fh = fopen_perhaps_with_pid(spesh_log, "w");
     spesh_disable = getenv("MVM_SPESH_DISABLE");
     if (!spesh_disable || strlen(spesh_disable) == 0) {
         instance->spesh_enabled = 1;
@@ -157,7 +212,7 @@ MVMInstance * MVM_vm_create_instance(void) {
         instance->jit_enabled = 1;
     jit_log = getenv("MVM_JIT_LOG");
     if (jit_log && strlen(jit_log))
-        instance->jit_log_fh = fopen(jit_log, "w");
+        instance->jit_log_fh = fopen_perhaps_with_pid(jit_log, "w");
     jit_bytecode_dir = getenv("MVM_JIT_BYTECODE_DIR");
     if (jit_bytecode_dir && strlen(jit_bytecode_dir)) {
         char *bytecode_map_name = MVM_malloc(strlen(jit_bytecode_dir) + strlen("/jit-map.txt") + 1);
@@ -171,7 +226,7 @@ MVMInstance * MVM_vm_create_instance(void) {
     /* Various kinds of debugging that can be enabled. */
     dynvar_log = getenv("MVM_DYNVAR_LOG");
     if (dynvar_log && strlen(dynvar_log)) {
-        instance->dynvar_log_fh = fopen(dynvar_log, "w");
+        instance->dynvar_log_fh = fopen_perhaps_with_pid(dynvar_log, "w");
         fprintf(instance->dynvar_log_fh, "+ x 0 0 0 0 0 %llu\n", uv_hrtime());
         fflush(instance->dynvar_log_fh);
         instance->dynvar_log_lasttime = uv_hrtime();
@@ -207,13 +262,16 @@ MVMInstance * MVM_vm_create_instance(void) {
 /* Set up some standard file handles. */
 static void setup_std_handles(MVMThreadContext *tc) {
     tc->instance->stdin_handle  = MVM_file_get_stdstream(tc, 0, 1);
-    MVM_gc_root_add_permanent(tc, (MVMCollectable **)&tc->instance->stdin_handle);
+    MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&tc->instance->stdin_handle,
+        "stdin handle");
 
     tc->instance->stdout_handle = MVM_file_get_stdstream(tc, 1, 0);
-    MVM_gc_root_add_permanent(tc, (MVMCollectable **)&tc->instance->stdout_handle);
+    MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&tc->instance->stdout_handle,
+        "stdout handle");
 
     tc->instance->stderr_handle = MVM_file_get_stdstream(tc, 2, 0);
-    MVM_gc_root_add_permanent(tc, (MVMCollectable **)&tc->instance->stderr_handle);
+    MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&tc->instance->stderr_handle,
+        "stderr handle");
 }
 
 /* This callback is passed to the interpreter code. It takes care of making
@@ -225,8 +283,6 @@ static void toplevel_initial_invoke(MVMThreadContext *tc, void *data) {
 
 /* Loads bytecode from the specified file name and runs it. */
 void MVM_vm_run_file(MVMInstance *instance, const char *filename) {
-    MVMStaticFrame *start_frame;
-
     /* Map the compilation unit into memory and dissect it. */
     MVMThreadContext *tc = instance->main_thread;
     MVMCompUnit      *cu = MVM_cu_map_from_file(tc, filename);
@@ -243,10 +299,8 @@ void MVM_vm_run_file(MVMInstance *instance, const char *filename) {
         }
     });
 
-    /* Run the frame marked main, or if there is none then fall back to the
-     * first frame. */
-    start_frame = cu->body.main_frame ? cu->body.main_frame : cu->body.frames[0];
-    MVM_interp_run(tc, toplevel_initial_invoke, start_frame);
+    /* Run the entry-point frame. */
+    MVM_interp_run(tc, toplevel_initial_invoke, cu->body.main_frame);
 }
 
 /* Loads bytecode from the specified file name and dumps it. */
@@ -255,8 +309,22 @@ void MVM_vm_dump_file(MVMInstance *instance, const char *filename) {
     MVMThreadContext *tc = instance->main_thread;
     MVMCompUnit      *cu = MVM_cu_map_from_file(tc, filename);
     char *dump = MVM_bytecode_dump(tc, cu);
+    size_t dumplen = strlen(dump);
+    int position = 0;
+
+    /* libuv already set up stdout to be nonblocking, but it can very well be
+     * we encounter EAGAIN (Resource temporarily unavailable), so we need to
+     * loop over our buffer, which can be quite big.
+     *
+     * The CORE.setting.moarvm has - as of writing this - about 32 megs of
+     * output from dumping.
+     */
+    while (position < dumplen) {
+        size_t written = write(1, dump + position, dumplen - position);
+        if (written > 0)
+            position += written;
+    }
 
-    printf("%s", dump);
     MVM_free(dump);
 }
 
@@ -284,6 +352,32 @@ void MVM_vm_exit(MVMInstance *instance) {
     exit(0);
 }
 
+static void cleanup_callsite_interns(MVMInstance *instance) {
+    int i;
+
+    for (i = 0; i < MVM_INTERN_ARITY_LIMIT; i++) {
+        int callsite_count = instance->callsite_interns->num_by_arity[i];
+        int j;
+
+        if (callsite_count) {
+            MVMCallsite **callsites = instance->callsite_interns->by_arity[i];
+
+            for (j = 0; j < callsite_count; j++) {
+                MVMCallsite *callsite = callsites[j];
+
+                if (MVM_callsite_is_common(callsite)) {
+                    continue;
+                }
+
+                MVM_callsite_destroy(callsite);
+            }
+
+            MVM_free(callsites);
+        }
+    }
+    MVM_free(instance->callsite_interns);
+}
+
 /* Destroys a VM instance. This must be called only from the main thread. It
  * should clear up all resources and free all memory; in practice, it falls
  * short of this goal at the moment. */
@@ -303,6 +397,7 @@ void MVM_vm_destroy_instance(MVMInstance *instance) {
     /* Clean up GC permanent roots related resources. */
     uv_mutex_destroy(&instance->mutex_permroots);
     MVM_free(instance->permroots);
+    MVM_free(instance->permroot_descriptions);
 
     /* Clean up Hash of HLLConfig. */
     uv_mutex_destroy(&instance->mutex_hllconfigs);
@@ -321,9 +416,10 @@ void MVM_vm_destroy_instance(MVMInstance *instance) {
     uv_mutex_destroy(&instance->mutex_extop_registry);
     MVM_HASH_DESTROY(hash_handle, MVMExtOpRegistry, instance->extop_registry);
 
-    /* Clean up Hash of all known serialization contexts. */
+    /* Clean up Hash of all known serialization contexts, along with list. */
     uv_mutex_destroy(&instance->mutex_sc_weakhash);
     MVM_HASH_DESTROY(hash_handle, MVMSerializationContextBody, instance->sc_weakhash);
+    MVM_free(instance->all_scs);
 
     /* Clean up Hash of filenames of compunits loaded from disk. */
     uv_mutex_destroy(&instance->mutex_loaded_compunits);
@@ -342,12 +438,36 @@ void MVM_vm_destroy_instance(MVMInstance *instance) {
     /* Clean up multi cache addition mutex. */
     uv_mutex_destroy(&instance->mutex_multi_cache_add);
 
+    /* Clean up interned callsites */
+    uv_mutex_destroy(&instance->mutex_callsite_interns);
+    cleanup_callsite_interns(instance);
+
+    /* Release this interpreter's hold on Unicode database */
+    MVM_unicode_release(instance->main_thread);
+
     /* Clean up spesh install mutex and close any log. */
     uv_mutex_destroy(&instance->mutex_spesh_install);
     if (instance->spesh_log_fh)
         fclose(instance->spesh_log_fh);
     if (instance->jit_log_fh)
         fclose(instance->jit_log_fh);
+    if (instance->dynvar_log_fh)
+        fclose(instance->dynvar_log_fh);
+
+    /* Clean up cross-thread-write-logging mutex */
+    uv_mutex_destroy(&instance->mutex_cross_thread_write_logging);
+
+    /* Clean up NFG. */
+    uv_mutex_destroy(&instance->nfg->update_mutex);
+    MVM_nfg_destroy(instance->main_thread);
+
+    /* Clean up fixed size allocator */
+    MVM_fixed_size_destroy(instance->fsa);
+
+    /* Clean up integer constant and string cache. */
+    uv_mutex_destroy(&instance->mutex_int_const_cache);
+    MVM_free(instance->int_const_cache);
+    MVM_free(instance->int_to_str_cache);
 
     /* Clean up event loop starting mutex. */
     uv_mutex_destroy(&instance->mutex_event_loop_start);
