@@ -59,6 +59,10 @@ MVMSpeshGraph * MVM_spesh_inline_try_get_graph(MVMThreadContext *tc, MVMSpeshGra
     if (target->body.sf == inliner->sf)
         return NULL;
 
+    /* Ensure they're from the same HLL. */
+    if (target->body.sf->body.cu->body.hll_config != inliner->sf->body.cu->body.hll_config)
+        return NULL;
+
     /* Ensure the candidate isn't still logging. */
     if (cand->sg)
         return NULL;
@@ -155,7 +159,7 @@ static void fix_coderef(MVMThreadContext *tc, MVMSpeshGraph *inliner,
 static void fix_str(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                     MVMSpeshGraph *inlinee, MVMSpeshOperand *to_fix) {
     to_fix->lit_str_idx = MVM_cu_string_add(tc, inliner->sf->body.cu,
-        inlinee->sf->body.cu->body.strings[to_fix->lit_str_idx]);
+        MVM_cu_string(tc, inlinee->sf->body.cu, to_fix->lit_str_idx));
 }
 static void fix_wval(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                      MVMSpeshGraph *inlinee, MVMSpeshIns *to_fix) {
@@ -363,8 +367,9 @@ static void merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
     inliner->inlines = inliner->num_inlines
         ? MVM_realloc(inliner->inlines, total_inlines * sizeof(MVMSpeshInline))
         : MVM_malloc(total_inlines * sizeof(MVMSpeshInline));
-    memcpy(inliner->inlines + inliner->num_inlines, inlinee->inlines,
-        inlinee->num_inlines * sizeof(MVMSpeshInline));
+    if (inlinee->num_inlines)
+        memcpy(inliner->inlines + inliner->num_inlines, inlinee->inlines,
+            inlinee->num_inlines * sizeof(MVMSpeshInline));
     for (i = inliner->num_inlines; i < total_inlines - 1; i++) {
         inliner->inlines[i].locals_start += inliner->num_locals;
         inliner->inlines[i].lexicals_start += inliner->num_lexicals;
@@ -401,26 +406,28 @@ static void merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
     inliner->num_inlines = total_inlines;
 
     /* Create/update per-specialization local and lexical type maps. */
-    if (!inliner->local_types) {
+    if (!inliner->local_types && inliner->num_locals) {
         MVMint32 local_types_size = inliner->num_locals * sizeof(MVMuint16);
         inliner->local_types = MVM_malloc(local_types_size);
         memcpy(inliner->local_types, inliner->sf->body.local_types, local_types_size);
     }
     inliner->local_types = MVM_realloc(inliner->local_types,
         (inliner->num_locals + inlinee->num_locals) * sizeof(MVMuint16));
-    memcpy(inliner->local_types + inliner->num_locals,
-        inlinee->local_types ? inlinee->local_types : inlinee->sf->body.local_types,
-        inlinee->num_locals * sizeof(MVMuint16));
-    if (!inliner->lexical_types) {
+    if (inlinee->num_locals)
+        memcpy(inliner->local_types + inliner->num_locals,
+            inlinee->local_types ? inlinee->local_types : inlinee->sf->body.local_types,
+            inlinee->num_locals * sizeof(MVMuint16));
+    if (!inliner->lexical_types && inliner->num_lexicals) {
         MVMint32 lexical_types_size = inliner->num_lexicals * sizeof(MVMuint16);
         inliner->lexical_types = MVM_malloc(lexical_types_size);
         memcpy(inliner->lexical_types, inliner->sf->body.lexical_types, lexical_types_size);
     }
     inliner->lexical_types = MVM_realloc(inliner->lexical_types,
         (inliner->num_lexicals + inlinee->num_lexicals) * sizeof(MVMuint16));
-    memcpy(inliner->lexical_types + inliner->num_lexicals,
-        inlinee->lexical_types ? inlinee->lexical_types : inlinee->sf->body.lexical_types,
-        inlinee->num_lexicals * sizeof(MVMuint16));
+    if (inlinee->num_lexicals)
+        memcpy(inliner->lexical_types + inliner->num_lexicals,
+            inlinee->lexical_types ? inlinee->lexical_types : inlinee->sf->body.lexical_types,
+            inlinee->num_lexicals * sizeof(MVMuint16));
 
     /* Merge handlers from inlinee. */
     if (inlinee->num_handlers) {
@@ -780,6 +787,39 @@ static void rewrite_args(MVMThreadContext *tc, MVMSpeshGraph *inliner,
         bb = bb->linear_next;
     }
 
+    {
+    MVMSpeshIns *arg_ins = call_info->prepargs_ins->next;
+    /* If there's some args that are not fetched by our inlinee,
+     * we have to kick them out, as arg_* ops are only valid between
+     * a prepargs and invoke_* op. */
+    while (arg_ins) {
+        MVMuint16    opcode = arg_ins->info->opcode;
+        MVMSpeshIns *next   = arg_ins->next;
+        switch (opcode) {
+            case MVM_OP_arg_i:
+            case MVM_OP_arg_n:
+            case MVM_OP_arg_s:
+            case MVM_OP_arg_o:
+                MVM_spesh_get_facts(tc, inliner, arg_ins->operands[1])->usages--;
+            case MVM_OP_argconst_i:
+            case MVM_OP_argconst_n:
+            case MVM_OP_argconst_s:
+                MVM_spesh_manipulate_delete_ins(tc, inliner, bb, arg_ins);
+                break;
+            case MVM_OP_set:
+                break;
+            case MVM_OP_invoke_i:
+            case MVM_OP_invoke_n:
+            case MVM_OP_invoke_s:
+            case MVM_OP_invoke_o:
+            case MVM_OP_invoke_v:
+            default:
+                next = NULL;
+        }
+        arg_ins = next;
+    }
+    }
+
     /* Delete the prepargs instruction. */
     MVM_spesh_manipulate_delete_ins(tc, inliner, invoke_bb, call_info->prepargs_ins);
 }
