@@ -291,7 +291,7 @@ MVMString * MVM_string_utf8_decode_strip_bom(MVMThreadContext *tc, const MVMObje
 
 /* Decodes using a decodestream. Decodes as far as it can with the input
  * buffers, or until a stopper is reached. */
-void MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
+MVMuint32 MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
                                   const MVMint32 *stopper_chars,
                                   MVMDecodeStreamSeparators *seps) {
     MVMint32 count = 0, total = 0;
@@ -302,15 +302,16 @@ void MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
     MVMDecodeStreamBytes *cur_bytes;
     MVMDecodeStreamBytes *last_accept_bytes = ds->bytes_head;
     MVMint32 last_accept_pos, ready, at_start;
+    MVMuint32 reached_stopper;
 
     /* If there's no buffers, we're done. */
     if (!ds->bytes_head)
-        return;
+        return 0;
     last_accept_pos = ds->bytes_head_pos;
 
     /* If we're asked for zero chars, also done. */
     if (stopper_chars && *stopper_chars == 0)
-        return;
+        return 1;
 
     /* Rough starting-size estimate is number of bytes in the head buffer. */
     bufsize = ds->bytes_head->length;
@@ -319,6 +320,7 @@ void MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
     /* Decode each of the buffers. */
     cur_bytes = ds->bytes_head;
     at_start = ds->abs_byte_pos == 0;
+    reached_stopper = 0;
     while (cur_bytes) {
         /* Process this buffer. */
         MVMint32  pos   = cur_bytes == ds->bytes_head ? ds->bytes_head_pos : 0;
@@ -358,10 +360,14 @@ void MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
                     }
                     buffer[count++] = g;
                     total++;
-                    if (stopper_chars && *stopper_chars == total)
+                    if (stopper_chars && *stopper_chars == total) {
+                        reached_stopper = 1;
                         goto done;
-                    if (MVM_string_decode_stream_maybe_sep(tc, seps, g))
+                    }
+                    if (MVM_string_decode_stream_maybe_sep(tc, seps, g)) {
+                        reached_stopper = 1;
                         goto done;
+                    }
                 }
                 break;
             }
@@ -383,11 +389,14 @@ void MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
         MVM_free(buffer);
     }
     MVM_string_decodestream_discard_to(tc, ds, last_accept_bytes, last_accept_pos);
+
+    return reached_stopper;
 }
 
 /* Encodes the specified string to UTF-8. */
 char * MVM_string_utf8_encode_substr(MVMThreadContext *tc,
-        MVMString *str, MVMuint64 *output_size, MVMint64 start, MVMint64 length, MVMString *replacement) {
+        MVMString *str, MVMuint64 *output_size, MVMint64 start, MVMint64 length,
+        MVMString *replacement, MVMint32 translate_newlines) {
     MVMuint8        *result;
     size_t           result_pos, result_limit;
     MVMCodepointIter ci;
@@ -403,7 +412,8 @@ char * MVM_string_utf8_encode_substr(MVMThreadContext *tc,
         MVM_exception_throw_adhoc(tc, "length out of range");
 
     if (replacement)
-        repl_bytes = MVM_string_utf8_encode_substr(tc, replacement, &repl_length, 0, -1, NULL);
+        repl_bytes = (MVMuint8 *) MVM_string_utf8_encode_substr(tc,
+            replacement, &repl_length, 0, -1, NULL, translate_newlines);
 
     /* Guesstimate that we'll be within 2 bytes for most chars most of the
      * time, and give ourselves 4 bytes breathing space. */
@@ -412,7 +422,7 @@ char * MVM_string_utf8_encode_substr(MVMThreadContext *tc,
     result_pos   = 0;
 
     /* Iterate the codepoints and encode them. */
-    MVM_string_ci_init(tc, &ci, str);
+    MVM_string_ci_init(tc, &ci, str, translate_newlines);
     while (MVM_string_ci_has_more(tc, &ci)) {
         MVMint32 bytes;
         MVMCodepoint cp = MVM_string_ci_get_codepoint(tc, &ci);
@@ -447,15 +457,17 @@ char * MVM_string_utf8_encode_substr(MVMThreadContext *tc,
 }
 
 /* Encodes the specified string to UTF-8. */
-char * MVM_string_utf8_encode(MVMThreadContext *tc, MVMString *str, MVMuint64 *output_size) {
-    return MVM_string_utf8_encode_substr(tc, str, output_size, 0, -1, NULL);
+char * MVM_string_utf8_encode(MVMThreadContext *tc, MVMString *str, MVMuint64 *output_size,
+        MVMint32 translate_newlines) {
+    return MVM_string_utf8_encode_substr(tc, str, output_size, 0, -1, NULL,
+        translate_newlines);
 }
 
 /* Encodes the specified string to a UTF-8 C string. */
 char * MVM_string_utf8_encode_C_string(MVMThreadContext *tc, MVMString *str) {
     MVMuint64 output_size;
     char * result;
-    char * utf8_string = MVM_string_utf8_encode(tc, str, &output_size);
+    char * utf8_string = MVM_string_utf8_encode(tc, str, &output_size, 0);
     /* this is almost always called from error-handling code. Don't care if it
      * contains embedded NULs. XXX TODO: Make sure all uses of this free what it returns */
     result = MVM_malloc(output_size + 1);
